package irnatura.offline;

import irnatura.offline.TempForwardIterator;
import java.util.TreeMap;

public class TempProfile
{
    // We don't expect this to be a legit temperature
    public static final int INVALID = -30000;

    private int m_min_temp;
    private int m_max_temp;
    private int m_cardinality; // may not be computed
    private int m_bitdepth;
    private int m_mean;         // may not be computed
    private int m_stdev;

    private int m_offset;
    private int m_scale;

    // Copy constructor
    public TempProfile(TempProfile rhs)
    {
        m_min_temp = rhs.m_min_temp;
        m_max_temp = rhs.m_max_temp;
        m_cardinality = rhs.m_cardinality;
        m_bitdepth = rhs.m_bitdepth;
        m_mean = rhs.m_mean;
        m_stdev = rhs.m_stdev;
        m_offset = rhs.m_offset;
        m_scale = rhs.m_scale;
    }

    // Compute from Histogram
    public TempProfile(TreeMap<Integer, Integer> histo)
    {
        m_min_temp = histo.firstKey();
        m_max_temp = histo.lastKey();
        m_cardinality = histo.size(); // For palette or any lossy mode
        m_mean = INVALID;   // Uncomputed
        m_stdev = INVALID;  // Uncomputed
        remainingStats();
    }

    // Compute from TempForwardIterator
    public TempProfile(TempForwardIterator it)
    {
        final int base = it.getTemp(0,0);
        int sum = 0; // sum of (t - base), to avoid overflow
        long sum2 = 0; // sum of squares
        m_min_temp = 30000;
        m_max_temp = -30000;
        for (int i = 0; i < it.getSize(); ++i) {
            int t = it.getTemp(i);
            int adjusted = t - base;
            sum += adjusted;
            sum2 += (long)adjusted * (long)adjusted;
            if (m_min_temp > t) m_min_temp = t;
            if (m_max_temp < t) m_max_temp = t;
        }
        int count = it.getSize();
        m_mean = Math.round((float)sum / (float)count) + base;
        m_stdev = (int)Math.round(
                        Math.sqrt((double)(count * sum2 - m_mean * m_mean))
                        / (double)count);
        m_cardinality = INVALID; // Uncomputed
        remainingStats();
    }

    // Recover from PNG file
    public TempProfile(String descr)
    {
        if (!fromString(descr)) {
            throw new UnsupportedOperationException("Description parse error");
        }
    }

    // Update min/max to encompass rhs's min/max.
    // When done, call remainingStats().
    public void updateMinMax(TempProfile rhs)
    {
        if (m_min_temp > rhs.m_min_temp)
            m_min_temp = rhs.m_min_temp;
        if (m_max_temp < rhs.m_max_temp)
            m_max_temp = rhs.m_max_temp;
        // remainingStats(); // TODO: uncomment for strict invariants
    }

    // After computing min and max and optionally num_elements and mean,
    // compute the remaining values.
    public void remainingStats()
    {
        int range = m_max_temp - m_min_temp + 1;
        m_bitdepth = 0;
        for (m_bitdepth = 4; m_bitdepth <= 16; m_bitdepth <<= 1) {
            final int ub = 2 << (m_bitdepth - 1);
            if (range < ub)
                break;
        }
        if (m_bitdepth > 16) {
            throw new java.lang.IndexOutOfBoundsException(
                    "Range too large: " + range);
        }
        // Compute scale and offset
        final int ub = 2 << (m_bitdepth - 1);
        m_scale = ub/range;
        m_offset = (ub - (range * m_scale)) / 2;
    }

    public int getMin() { return m_min_temp; }
    public int getMax() { return m_max_temp; }
    public int getCardinality() { return m_cardinality; } // or INVALID
    public int getBitDepth() { return m_bitdepth; }
    public int getMean() { return m_mean; } // or INVALID
    public int getStdev() { return m_stdev; } // or INVALID

    // Try to adjust stats, assuming oldval is replaced with newval.
    // For now, just min and max.
    // Could do mean, but useful?
    // After all has settled, must call remainingStats().
    public void tweak(int oldval, int newval)
    {
        if (m_min_temp > newval) m_min_temp = newval;
        if (m_max_temp < newval) m_max_temp = newval;
    }

    public int getGrayValue(int temp)
    {
        int zerobase = temp - m_min_temp;
        int scaled = zerobase * m_scale;
        int shifted = scaled + m_offset;
        return shifted;
    }

    public int fromGrayValue(int val)
    {
        int scaled = val - m_offset;
        int zerobase = scaled / m_scale; // assert no remainder
        if (scaled % m_scale != 0) {
            throw new java.lang.ArithmeticException(
                    "Value " + val + " doesn't divide by scale=" + m_scale);
        }
        int temp = zerobase + m_min_temp;
        return temp;
    }

    public String toString()
    {
        return "min=" + m_min_temp + ";"
            + "max=" + m_max_temp + ";"
            + ((m_cardinality == INVALID)
                ? "" : ("card=" + m_cardinality + ";"))
            + "bitdepth=" + m_bitdepth + ";"
            + ((m_mean == INVALID)
                ? "" : ("mean=" + m_mean + ";"))
            + ((m_stdev == INVALID)
                ? "" : ("stdev=" + m_stdev + ";"))
            + "offset=" + m_offset + ";"
            + "scale=" + m_scale + ";"
            ;
    }

    private boolean fromString(String s)
    {
        m_min_temp = INVALID;
        m_max_temp = INVALID;
        m_cardinality = INVALID;
        m_bitdepth = INVALID;
        m_mean = INVALID;
        m_stdev = INVALID;
        m_offset = INVALID;
        m_scale = INVALID;
        String [] kvs = s.split(";");
        for (String kev : kvs) {
            String [] kv = kev.split("=", 2);
            if (kv.length != 2)
                continue; // nonfatal?
            if (kv[0].equals("min"))
                m_min_temp = Integer.parseInt(kv[1]);
            else if (kv[0].equals("max"))
                m_max_temp = Integer.parseInt(kv[1]);
            else if (kv[0].equals("card"))
                m_cardinality = Integer.parseInt(kv[1]);
            else if (kv[0].equals("bitdepth"))
                m_bitdepth = Integer.parseInt(kv[1]);
            else if (kv[0].equals("mean"))
                m_mean = Integer.parseInt(kv[1]);
            else if (kv[0].equals("stdev"))
                m_stdev = Integer.parseInt(kv[1]);
            else if (kv[0].equals("offset"))
                m_offset = Integer.parseInt(kv[1]);
            else if (kv[0].equals("scale"))
                m_scale = Integer.parseInt(kv[1]);
        }
        return m_min_temp != INVALID
            && m_max_temp != INVALID
         // && m_cardinality != INVALID     // Allow this to be invalid
            && m_bitdepth != INVALID
         // && m_mean != INVALID            // Allow this to be invalid
         // && m_stdev != INVALID           // Allow this to be invalid
            && m_offset != INVALID
            && m_scale != INVALID
            ;
    }
}
